import pytest

from npe2 import read, read_get_reader, write, write_get_writer
from npe2.types import FullLayerData

SAMPLE_PLUGIN_NAME = "my-plugin"

SAMPLE_PLUGIN_NAME = "my-plugin"


def test_read(uses_sample_plugin):
    assert read(["some.fzzy"], stack=False) == [(None,)]


def test_read_with_plugin(uses_sample_plugin):
    # no such plugin name.... but skips over the sample plugin
    with pytest.raises(ValueError):
        read(["some.fzzy"], plugin_name="nope", stack=False)


def test_read_return_reader(uses_sample_plugin):
    data, reader = read_get_reader("some.fzzy")
    assert data == [(None,)]
    assert reader.command == f"{SAMPLE_PLUGIN_NAME}.some_reader"


def test_read_return_reader_with_stack(uses_sample_plugin):
    data, reader = read_get_reader(["some.fzzy"], stack=True)
    assert data == [(None,)]
    assert reader.command == f"{SAMPLE_PLUGIN_NAME}.some_reader"


def test_read_list(uses_sample_plugin):
    data, reader = read_get_reader(["some.fzzy", "other.fzzy"])
    assert data == [(None,)]
    assert reader.command == f"{SAMPLE_PLUGIN_NAME}.some_reader"


null_image: FullLayerData = ([], {}, "image")


def test_writer_exec(uses_sample_plugin):
    # the sample writer knows how to handle two image layers
    result = write("test.tif", [null_image, null_image])
    assert result == ["test.tif"]

    result, contrib = write_get_writer("test.tif", [null_image, null_image])
    assert result == ["test.tif"]
    assert contrib.command == f"{SAMPLE_PLUGIN_NAME}.my_writer"


@pytest.mark.parametrize("layer_data", [[null_image, null_image], []])
def test_writer_exec_fails(layer_data, uses_sample_plugin):
    # the sample writer doesn't accept no extension
    with pytest.raises(ValueError):
        write("test/path", layer_data)


def test_writer_exec_fails2(uses_sample_plugin):
    # the sample writer doesn't accept 5 images
    with pytest.raises(ValueError):
        write("test.tif", [null_image, null_image, null_image, null_image, null_image])


def test_writer_single_layer_api_exec(uses_sample_plugin):
    # This writer doesn't do anything but type check.
    paths = write("test/path", [([], {}, "labels")])
    assert len(paths) == 1



# flake8: noqa
import ast
from pathlib import Path
from types import ModuleType
from typing import Any, Callable, Dict, List, TypeVar, Union

from _typeshed import Incomplete
from pydantic import BaseModel as BaseModel

from .manifest import contributions as contributions

T = TypeVar("T", bound=Callable[..., Any])

def reader(
    *,
    id: str,
    title: str,
    filename_patterns: List[str],
    accepts_directories: bool = False,
    ensure_args_valid: bool = False,
) -> Callable[[T], T]:
    """Mark a function as a reader contribution"""

def writer(
    *,
    id: str,
    title: str,
    layer_types: List[str],
    filename_extensions: List[str] = [],
    display_name: str = "",
    ensure_args_valid: bool = False,
) -> Callable[[T], T]:
    """Mark function as a writer contribution"""

def widget(
    *,
    id: str,
    title: str,
    display_name: str,
    autogenerate: bool = False,
    ensure_args_valid: bool = False,
) -> Callable[[T], T]:
    """Mark a function as a widget contribution"""

def sample_data_generator(
    *,
    id: str,
    title: str,
    key: str,
    display_name: str,
    ensure_args_valid: bool = False,
) -> Callable[[T], T]:
    """Mark a function as a sample data generator contribution"""

def on_activate(func):
    """Mark a function to be called when a plugin is activated."""

def on_deactivate(func):
    """Mark a function to be called when a plugin is deactivated."""

class PluginModuleVisitor(ast.NodeVisitor):
    """AST visitor to find all contributions in a module."""

    plugin_name: str
    module_name: str
    contribution_points: Dict[str, List[dict]]
    def __init__(self, plugin_name: str, module_name: str) -> None: ...
    def visit_Import(self, node: ast.Import) -> Any: ...
    def visit_ImportFrom(self, node: ast.ImportFrom) -> Any: ...
    def visit_FunctionDef(self, node: ast.FunctionDef) -> Any: ...
    def visit_ClassDef(self, node: ast.ClassDef) -> Any: ...

def visit(
    path: Union[ModuleType, str, Path], plugin_name: str, module_name: str = ...
) -> contributions.ContributionPoints:
    """Visit a module and extract contribution points."""



from typing import Dict, List, Optional

from pydantic import BaseModel, Field

from ._commands import CommandContribution
from ._menus import MenuItem
from ._readers import ReaderContribution
from ._sample_data import SampleDataContribution, SampleDataGenerator, SampleDataURI
from ._submenu import SubmenuContribution
from ._themes import ThemeContribution
from ._widgets import WidgetContribution
from ._writers import WriterContribution

__all__ = [
    "ContributionPoints",
    "CommandContribution",
    "MenuItem",
    "ReaderContribution",
    "SampleDataContribution",
    "SubmenuContribution",
    "ThemeContribution",
    "WidgetContribution",
    "WriterContribution",
    "SampleDataGenerator",
    "SampleDataURI",
]


class ContributionPoints(BaseModel):
    commands: Optional[List[CommandContribution]]
    readers: Optional[List[ReaderContribution]]
    writers: Optional[List[WriterContribution]]
    widgets: Optional[List[WidgetContribution]]
    sample_data: Optional[List[SampleDataContribution]]
    themes: Optional[List[ThemeContribution]]

    menus: Dict[str, List[MenuItem]] = Field(default_factory=dict, hide_docs=True)
    submenus: Optional[List[SubmenuContribution]] = Field(None, hide_docs=True)

    # configuration: Optional[JsonSchemaObject]
    # keybindings: Optional[List[KeyBindingContribution]]

    class Config:
        docs_exclude = {"menus", "submenus"}



from npe2 import PackageMetadata


def test_package_metadata_version():
    """Test that we intelligently pick the min required metadata version"""
    assert PackageMetadata(name="test", version="1.0").metadata_version == "1.0"
    pm2 = PackageMetadata(name="test", version="1.0", maintainer="bob")
    assert pm2.metadata_version == "1.2"
    pm3 = PackageMetadata(
        name="test",
        version="1.0",
        maintainer="bob",
        description_content_type="text/markdown",
    )
    assert pm3.metadata_version == "2.1"


def test_hashable():
    hash(PackageMetadata(name="test", version="1.0"))



import os
from unittest.mock import patch

import pytest

from npe2 import fetch_manifest
from npe2._fetch import (
    _fetch_manifest_with_full_install,
    get_hub_plugin,
    get_hub_plugins,
    get_pypi_url,
)
from npe2.manifest._npe1_adapter import NPE1Adapter


def test_npe2_from_pypi_wheel():
    mf = fetch_manifest("napari-omero")
    assert mf.name == "napari-omero"
    assert mf.contributions


@pytest.mark.parametrize("version", [None, "0.1.0"])
@pytest.mark.parametrize("packagetype", ["sdist", "bdist_wheel", None])
def test_get_pypi_url(version, packagetype):
    assert "npe2" in get_pypi_url("npe2", version=version, packagetype=packagetype)


def test_from_pypi_wheel_bdist_missing():
    error = KeyError("No bdist_wheel releases found")
    p2 = patch("npe2._fetch.get_pypi_url", side_effect=error)
    with patch("npe2._fetch._fetch_manifest_with_full_install") as mock_fetch, p2:
        fetch_manifest("my-package")
    mock_fetch.assert_called_once_with("my-package", version=None)


@pytest.mark.skipif(not os.getenv("CI"), reason="slow, only run on CI")
def test_fetch_manifest_with_full_install():
    # TODO: slowest of the tests ... would be nice to provide a local mock
    mf = _fetch_manifest_with_full_install("napari-ndtiffs")
    assert isinstance(mf, NPE1Adapter)
    assert mf.name == "napari-ndtiffs"
    assert mf.contributions


def test_get_hub_plugins():
    plugins = get_hub_plugins()
    assert "napari-svg" in plugins


def test_get_hub_plugin():
    info = get_hub_plugin("napari-svg")
    assert info["name"] == "napari-svg"



import ast
import contextlib
import inspect
from inspect import Parameter, Signature
from pathlib import Path
from types import ModuleType
from typing import Any, Callable, Dict, List, Sequence, Tuple, Type, TypeVar, Union

from pydantic import BaseModel

from .manifest import contributions

__all__ = [
    "on_activate",
    "on_deactivate",
    "PluginModuleVisitor",
    "reader",
    "sample_data_generator",
    "visit",
    "widget",
    "writer",
]

T = TypeVar("T", bound=Callable[..., Any])
_COMMAND_PARAMS = inspect.signature(contributions.CommandContribution).parameters
CONTRIB_MAP: Dict[str, Tuple[Type[BaseModel], str]] = {
    "writer": (contributions.WriterContribution, "writers"),
    "reader": (contributions.ReaderContribution, "readers"),
    "sample_data_generator": (contributions.SampleDataGenerator, "sample_data"),
    "widget": (contributions.WidgetContribution, "widgets"),
}
CHECK_ARGS_PARAM = "ensure_args_valid"


def _build_decorator(contrib: Type[BaseModel]) -> Callable:
    """Create a decorator (e.g. `@implements.reader`) to mark an object as a contrib.

    Parameters
    ----------
    contrib : Type[BaseModel]
        The type of contribution this object implements.
    """
    # build a signature based on the fields in this contribution type, mixed with
    # the fields in the CommandContribution
    contribs: Sequence[Type[BaseModel]] = (contributions.CommandContribution, contrib)
    params: List[Parameter] = []
    for contrib in contribs:
        # iterate over the fields in the contribution types
        for field in contrib.__fields__.values():
            # we don't need python_name (since that will be gleaned from the function
            # we're decorating) ... and we don't need `command`, since that will just
            # be a string pointing to the contributions.commands entry that we are
            # creating here.
            if field.name not in {"python_name", "command"}:
                # ensure that required fields raise a TypeError if they are not provided
                default = Parameter.empty if field.required else field.get_default()
                # create the parameter and add it to the signature.
                param = Parameter(
                    field.name,
                    Parameter.KEYWORD_ONLY,
                    default=default,
                    annotation=field.outer_type_ or field.type_,
                )
                params.append(param)

    # add one more parameter to control whether the arguments in the decorator itself
    # are validated at runtime
    params.append(
        Parameter(
            CHECK_ARGS_PARAM,
            kind=Parameter.KEYWORD_ONLY,
            default=False,
            annotation=bool,
        )
    )

    signature = Signature(parameters=params, return_annotation=Callable[[T], T])

    # creates the actual `@npe2.implements.something` decorator
    # this just stores the parameters for the corresponding contribution type
    # as attributes on the function being decorated.
    def _deco(**kwargs) -> Callable[[T], T]:
        def _store_attrs(func: T) -> T:
            # If requested, assert that we've satisfied the signature when
            # the decorator is invoked at runtime.
            # TODO: improve error message to provide context
            if kwargs.pop(CHECK_ARGS_PARAM, False):
                signature.bind(**kwargs)

            # TODO: check if it's already there and assert the same id
            # store these attributes on the function
            with contextlib.suppress(AttributeError):
                setattr(func, f"_npe2_{contrib.__name__}", kwargs)

            # return the original decorated function
            return func

        return _store_attrs

    # set the signature and return the decorator
    setattr(_deco, "__signature__", signature)
    return _deco


# builds decorators for each of the contribution types that are essentially just
# pointers to some command.
reader = _build_decorator(contributions.ReaderContribution)
writer = _build_decorator(contributions.WriterContribution)
widget = _build_decorator(contributions.WidgetContribution)
sample_data_generator = _build_decorator(contributions.SampleDataGenerator)


def on_activate(func):
    """Mark a function to be called when a plugin is activated."""
    setattr(func, "npe2_on_activate", True)
    return func


def on_deactivate(func):
    """Mark a function to be called when a plugin is deactivated."""
    setattr(func, "npe2_on_deactivate", True)
    return func


class PluginModuleVisitor(ast.NodeVisitor):
    """AST visitor to find all contributions in a module.

    This visitor will find all the contributions in a module and store them in
    `contribution_points`.  It works as follows:

    1. Visit all `Import` and `ImportFrom` nodes in the module, storing their import
       names in `_names` so we can look them up later.  For example, if the module
       had the line `from npe2 import implements as impls` at the top, then the
       `visit_ImportFrom` method would add the entry:
       `self._names['impls'] = 'npe2.implements'`
       This way, we know that an `@impls` found later in the module is referring to
       `npe2.implements`.
    2. Visit all `FunctionDef` and `ClassDef` nodes in the module and check their
       decorators with `_find_decorators`
    3. In `_find_decorators` we check to see if the name of any of the decorators
       resolves to the something from this module (i.e., if it's being decorated with
       something from `npe2.implements`).  If it is, then we call `_store_contrib`...
    4. `_store_contrib` first calls `_store_command` which does the job of storing the
       fully-resolved `python_name` for the function being decorated, and creates a
       CommandContribution. `_store_contrib` will then create the appropriate
       contribution type (e.g. `npe2.implements.reader` will instantiate a
       `ReaderContribution`), set its `command` entry to the `id` of the just-created
       `CommandContribution`, then store it in `contribution_points`.
    5. When the visitor is finished, we can create an instance of `ContributionPoints`
       using `ContributionPoints(**visitor.contribution_points)`, then add it to a
       manifest.
    """

    def __init__(self, plugin_name: str, module_name: str) -> None:
        super().__init__()
        self.plugin_name = plugin_name
        self.module_name = module_name
        self.contribution_points: Dict[str, List[dict]] = {}
        self._names: Dict[str, str] = {}

    def visit_Import(self, node: ast.Import) -> Any:  # noqa: D102
        # https://docs.python.org/3/library/ast.html#ast.Import
        for alias in node.names:
            self._names[alias.asname or alias.name] = alias.name
        return super().generic_visit(node)

    def visit_ImportFrom(self, node: ast.ImportFrom) -> Any:  # noqa: D102
        # https://docs.python.org/3/library/ast.html#ast.ImportFrom
        for alias in node.names:
            self._names[alias.asname or alias.name] = f"{node.module}.{alias.name}"
        return super().generic_visit(node)

    def visit_FunctionDef(self, node: ast.FunctionDef) -> Any:  # noqa: D102
        # https://docs.python.org/3/library/ast.html#ast.FunctionDef
        self._find_decorators(node)

    def visit_ClassDef(self, node: ast.ClassDef) -> Any:  # noqa: D102
        self._find_decorators(node)

    def _find_decorators(self, node: Union[ast.ClassDef, ast.FunctionDef]):
        # for each in the decorator list ...
        for call in node.decorator_list:
            # https://docs.python.org/3/library/ast.html#ast.Call
            # if the function is an attribute ...
            # (e.g in `@npe2.implements.reader`, `reader` is an attribute of
            # implements, which is an attribute of npe2)
            if not isinstance(call, ast.Call):
                continue
            if isinstance(call.func, ast.Attribute):
                # then go up the chain of attributes until we get to a root Name
                val = call.func.value
                _names = []
                while isinstance(val, ast.Attribute):
                    _names.append(val.attr)
                    val = val.value
                if isinstance(val, ast.Name):
                    _names.append(val.id)
                    # finally, we can build the fully resolved call name of the
                    # decorator (e.g. `@npe2.implements`, or `@implements`)
                    call_name = ".".join(reversed(_names))
                    # we then check the `_names` we gathered above to resolve these
                    # call names to their fully qualified names (e.g. `implements`
                    # would resolve to `npe2.implements` if the name `implements` was
                    # imported from `npe2`.)
                    # If the name resolves to the name of this module,
                    # then we have a hit! Store the contribution.
                    if self._names.get(call_name) == __name__:
                        kwargs = self._keywords_to_kwargs(call.keywords)
                        self._store_contrib(call.func.attr, node.name, kwargs)
            elif isinstance(call.func, ast.Name):
                # if the function is just a direct name (e.g. `@reader`)
                # then we can just see if the name points to something imported from
                # this module.
                if self._names.get(call.func.id, "").startswith(__name__):
                    kwargs = self._keywords_to_kwargs(call.keywords)
                    self._store_contrib(call.func.id, node.name, kwargs)
        return super().generic_visit(node)

    def _keywords_to_kwargs(self, keywords: List[ast.keyword]) -> Dict[str, Any]:
        return {str(k.arg): ast.literal_eval(k.value) for k in keywords}

    def _store_contrib(self, contrib_type: str, name: str, kwargs: Dict[str, Any]):
        kwargs.pop(CHECK_ARGS_PARAM, None)
        ContribClass, contrib_name = CONTRIB_MAP[contrib_type]
        contrib = ContribClass(**self._store_command(name, kwargs))
        existing: List[dict] = self.contribution_points.setdefault(contrib_name, [])
        existing.append(contrib.dict(exclude_unset=True))

    def _store_command(self, name: str, kwargs: Dict[str, Any]) -> Dict[str, Any]:
        cmd_kwargs = {k: kwargs.pop(k) for k in list(kwargs) if k in _COMMAND_PARAMS}
        cmd_kwargs["python_name"] = self._qualified_pyname(name)
        cmd = contributions.CommandContribution(**cmd_kwargs)
        if cmd.id.startswith(self.plugin_name):
            n = len(self.plugin_name)
            cmd.id = cmd.id[n:]
        cmd.id = f"{self.plugin_name}.{cmd.id.lstrip('.')}"
        cmd_contribs: List[dict] = self.contribution_points.setdefault("commands", [])
        cmd_contribs.append(cmd.dict(exclude_unset=True))
        kwargs["command"] = cmd.id
        return kwargs

    def _qualified_pyname(self, obj_name: str) -> str:
        return f"{self.module_name}:{obj_name}"


def visit(
    path: Union[ModuleType, str, Path], plugin_name: str, module_name: str = ""
) -> contributions.ContributionPoints:
    """Visit a module and extract contribution points.

    Parameters
    ----------
    path : Union[ModuleType, str, Path]
        Either a path to a Python module, a module object, or a string
    plugin_name : str
        Name of the plugin
    module_name : str
        Module name, by default ""

    Returns
    -------
    ContributionPoints
        ContributionPoints discovered in the module.
    """
    if isinstance(path, ModuleType):
        assert path.__file__
        assert path.__name__
        module_name = path.__name__
        path = path.__file__

    visitor = PluginModuleVisitor(plugin_name, module_name=module_name)
    visitor.visit(ast.parse(Path(path).read_text()))
    if "commands" in visitor.contribution_points:
        compress = {tuple(i.items()) for i in visitor.contribution_points["commands"]}
        visitor.contribution_points["commands"] = [dict(i) for i in compress]
    return contributions.ContributionPoints(**visitor.contribution_points)



"""
JsonSchemaObject class below vendored/modified from datamodel-code-generator

MIT License

Copyright (c) 2019 Koudai Aono

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

"""
from enum import Enum
from typing import Any, Dict, List, Literal, Optional, Set, Type, Union

from pydantic import BaseModel, Extra, Field, root_validator, validator

JsonType = Union[
    Literal["string"],
    Literal["number"],
    Literal["integer"],
    Literal["object"],
    Literal["array"],
    Literal["boolean"],
    Literal["null"],
]

_python_equivalent: Dict[Optional[str], Type] = {
    "string": str,
    "number": float,
    "integer": int,
    "object": dict,
    "array": list,
    "boolean": bool,
    "null": type(None),
    None: object,
}


class JSONReference(Enum):
    LOCAL = "LOCAL"
    REMOTE = "REMOTE"
    URL = "URL"


_CONSTRAINED_FIELDS = {
    "exclusiveMinimum",
    "minimum",
    "exclusiveMaximum",
    "maximum",
    "multipleOf",
    "minItems",
    "maxItems",
    "minLength",
    "maxLength",
    "pattern",
}


class JsonSchemaObject(BaseModel):
    items: Union[List["JsonSchemaObject"], "JsonSchemaObject", None]
    uniqueItem: Optional[bool]
    type: Union[JsonType, List[JsonType], None]
    format: Optional[str]
    pattern: Optional[str]
    minLength: Optional[int]
    maxLength: Optional[int]
    minimum: Optional[float]
    maximum: Optional[float]
    minItems: Optional[int]
    maxItems: Optional[int]
    multipleOf: Optional[float]
    exclusiveMaximum: Union[float, bool, None]
    exclusiveMinimum: Union[float, bool, None]
    additionalProperties: Union["JsonSchemaObject", bool, None]
    oneOf: List["JsonSchemaObject"] = []
    anyOf: List["JsonSchemaObject"] = []
    allOf: List["JsonSchemaObject"] = []
    enum: List[Any] = []
    enum_descriptions: List[str] = Field(
        default_factory=list, description="provides descriptive text for each enum."
    )
    markdown_enum_descriptions: Optional[List[str]] = Field(
        default_factory=list,
        description="If you use markdown_enum_descriptions instead of "
        "enum_descriptions, your descriptions will be rendered as Markdown",
    )
    writeOnly: Optional[bool]
    properties: Optional[Dict[str, "JsonSchemaObject"]]
    required: List[str] = []
    nullable: Optional[bool] = False
    x_enum_varnames: List[str] = Field(default_factory=list)
    description: Optional[str]
    markdown_description: Optional[str] = Field(
        None,
        description="If you use markdown_description instead of description, your "
        "setting description will be rendered as Markdown in the settings UI.",
    )

    deprecation_message: Optional[str] = Field(
        None,
        description="If you set deprecationMessage, the setting will get a warning "
        "underline with your specified message. It won't show up in the settings "
        "UI unless it is configured by the user.",
    )
    markdown_deprecation_message: Optional[str] = Field(
        None,
        description="If you set markdown_deprecation_message, the setting will get a "
        "warning underline with your specified message. It won't show up in the "
        "settings UI unless it is configured by the user.",
    )
    # NOTE:
    # If you set both properties, deprecation_message will be shown in the hover and
    # the problems view, and markdown_deprecation_message will be rendered as markdown
    # in the settings UI.
    title: Optional[str]
    example: Any
    examples: Any
    default: Any
    id: Optional[str] = Field(default=None)
    custom_type_path: Optional[str] = Field(default=None)
    _raw: Dict[str, Any]

    @root_validator(pre=True)
    def validate_exclusive_maximum_and_exclusive_minimum(
        cls, values: Dict[str, Any]
    ) -> Any:
        if "$ref" in values:
            # ref not supported
            values.pop("$ref")

        exclusive_maximum: Union[float, bool, None] = values.get("exclusiveMaximum")
        exclusive_minimum: Union[float, bool, None] = values.get("exclusiveMinimum")

        if exclusive_maximum is True:
            values["exclusiveMaximum"] = values["maximum"]
            del values["maximum"]
        elif exclusive_maximum is False:
            del values["exclusiveMaximum"]
        if exclusive_minimum is True:
            values["exclusiveMinimum"] = values["minimum"]
            del values["minimum"]
        elif exclusive_minimum is False:
            del values["exclusiveMinimum"]

        if "type" not in values and "properties" in values:
            values["type"] = "object"
        return values

    class Config:
        extra = Extra.forbid
        arbitrary_types_allowed = True
        underscore_attrs_are_private = True

    def __init__(self, **data: Any) -> None:
        super().__init__(**data)
        self._raw = data

    @property
    def extras(self) -> Dict[str, Any]:
        return {k: v for k, v in self._raw.items() if k not in EXCLUDE_FIELD_KEYS}

    @property
    def is_object(self) -> bool:
        return (
            self.properties is not None
            or self.type == "object"
            and not self.allOf
            and not self.oneOf
            and not self.anyOf
            and not getattr(self, "ref", False)
        )

    @property
    def python_type(self) -> Union[Type, List[Type]]:
        if isinstance(self.type, list):
            return [_python_equivalent[t] for t in self.type]
        else:
            return _python_equivalent[self.type]

    @property
    def is_array(self) -> bool:
        return self.items is not None or self.type == "array"

    @validator("items", pre=True)
    def validate_items(cls, values: Any) -> Any:
        # this condition expects empty dict
        return values or None

    @property
    def has_default(self) -> bool:
        return "default" in self.__fields_set__

    @property
    def has_constraint(self) -> bool:
        return bool(_CONSTRAINED_FIELDS & self.__fields_set__)

    # refs not supported
    # ref: Optional[str] = Field(default=None, alias="$ref")

    # @validator("ref")
    # def validate_ref(cls, value: Any) -> Any:
    #     if isinstance(value, str) and "#" in value:
    #         if value.endswith("#/"):
    #             return value[:-1]
    #         elif "#/" in value or value[0] == "#" or value[-1] == "#":
    #             return value
    #         return value.replace("#", "#/")
    #     return value

    # @property
    # def ref_object_name(self) -> Optional[str]:
    #     return self.ref.rsplit("/", 1)[-1] if self.ref else None

    # @property
    # def ref_type(self) -> Optional[JSONReference]:
    #     if self.ref:
    #         return get_ref_type(self.ref)
    #     return None


# def is_url(ref: str) -> bool:
#     return ref.startswith(("https://", "http://"))


# @lru_cache()
# def get_ref_type(ref: str) -> JSONReference:
#     if ref[0] == "#":
#         return JSONReference.LOCAL
#     elif is_url(ref):
#         return JSONReference.URL
#     return JSONReference.REMOTE


JsonSchemaObject.update_forward_refs()

DEFAULT_FIELD_KEYS: Set[str] = {
    "example",
    "examples",
    "description",
    "title",
}

EXCLUDE_FIELD_KEYS = (set(JsonSchemaObject.__fields__) - DEFAULT_FIELD_KEYS) | {
    "$id",
    "$ref",
}



from __future__ import annotations

import json
import site
import sys
import tempfile
import warnings
from contextlib import contextmanager
from functools import lru_cache
from importlib import metadata
from io import BytesIO
from logging import getLogger
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, Iterator, List, Optional
from urllib.request import urlopen
from zipfile import ZipFile

from build.env import IsolatedEnvBuilder

if TYPE_CHECKING:
    import build.env

    from npe2.manifest.schema import PluginManifest

logger = getLogger(__name__)

NPE1_ENTRY_POINT = "napari.plugin"
NPE2_ENTRY_POINT = "napari.manifest"
__all__ = [
    "fetch_manifest",
    "get_pypi_url",
    "isolated_plugin_env",
    "get_hub_plugin",
    "get_hub_plugins",
]


def fetch_manifest(package: str, version: Optional[str] = None) -> PluginManifest:
    """Fetch a manifest for a pip specifier (package name, possibly with version).

    Parameters
    ----------
    package : str
        package name
    version : Optional[str]
        package version, by default, latest version.

    Returns
    -------
    PluginManifest
        Plugin manifest for package `specifier`.
    """
    from npe2 import PluginManifest
    from npe2.manifest import PackageMetadata

    has_npe1 = False

    # first just grab the wheel from pypi with no dependencies
    try:
        with _tmp_pypi_wheel_download(package, version) as td:
            # create a PathDistribution from the dist-info directory in the wheel
            dist = metadata.PathDistribution(next(Path(td).glob("*.dist-info")))

            for ep in dist.entry_points:
                # if we find an npe2 entry point, we can just use
                # PathDistribution.locate_file to get the file.
                if ep.group == NPE2_ENTRY_POINT:
                    logger.debug("pypi wheel has npe2 entry point.")
                    # python 3.8 fallbacks
                    match = ep.pattern.match(ep.value)
                    assert match
                    module: str = match.groupdict()["module"]
                    attr: str = match.groupdict()["attr"]

                    mf_file = dist.locate_file(Path(module) / attr)
                    mf = PluginManifest.from_file(str(mf_file))
                    # manually add the package metadata from our distribution object.
                    mf.package_metadata = PackageMetadata.from_dist_metadata(
                        dist.metadata
                    )
                    return mf
                elif ep.group == NPE1_ENTRY_POINT:
                    has_npe1 = True  # pragma: no cover
    except KeyError as e:
        if "No bdist_wheel releases found" not in str(e):
            raise  # pragma: no cover
        has_npe1 = True

    if not has_npe1:
        raise ValueError(f"{package} had no napari entry points.")  # pragma: no cover

    logger.debug("falling back to npe1")
    return _fetch_manifest_with_full_install(package, version=version)


@lru_cache
def _pypi_info(package: str) -> dict:
    with urlopen(f"https://pypi.org/pypi/{package}/json") as f:
        return json.load(f)


def get_pypi_url(
    package: str, version: Optional[str] = None, packagetype: Optional[str] = None
) -> str:
    """Get URL for a package on PyPI.

    Parameters
    ----------
    package : str
        package name
    version : str, optional
        package version, by default, latest version.
    packagetype : str, optional
        one of `'sdist'`, `'bdist_wheel'`, or `None`,
        by default `None`, which means 'bdist_wheel' will be tried first, then 'sdist'

    Returns
    -------
    str
        URL to download the package.

    Raises
    ------
    ValueError
        If packagetype is not one of 'sdist', 'bdist_wheel', or if version is specified
        and does not match any available version.
    KeyError
        If packagetype is specified and no package of that type is available.
    """
    if packagetype not in {"sdist", "bdist_wheel", None}:
        raise ValueError(  # pragma: no cover
            f"Invalid packagetype: {packagetype}, must be one of sdist, bdist_wheel"
        )

    data = _pypi_info(package)
    if version:
        version = version.lstrip("v")
        try:
            _releases: List[dict] = data["releases"][version]
        except KeyError as e:  # pragma: no cover
            raise ValueError(f"{package} does not have version {version}") from e
    else:
        _releases = data["urls"]

    releases = {d.get("packagetype"): d for d in _releases}
    if packagetype:
        if packagetype not in releases:  # pragma: no cover
            version = version or "latest"
            raise KeyError(f'No {packagetype} releases found for version "{version}"')
        return releases[packagetype]["url"]
    return (releases.get("bdist_wheel") or releases["sdist"])["url"]


@contextmanager
def _tmp_pypi_wheel_download(
    package: str, version: Optional[str] = None
) -> Iterator[Path]:
    url = get_pypi_url(package, version=version, packagetype="bdist_wheel")
    logger.debug(f"downloading wheel for {package} {version or ''}")
    with tempfile.TemporaryDirectory() as td, urlopen(url) as f:
        with ZipFile(BytesIO(f.read())) as zf:
            zf.extractall(td)
            yield Path(td)


@contextmanager
def isolated_plugin_env(
    package: str,
    version: Optional[str] = None,
    validate_npe1_imports: bool = True,
    install_napari_if_necessary: bool = True,
) -> Iterator[build.env.IsolatedEnv]:
    """Isolated env context with a plugin installed.

    The site-packages folder of the env is added to sys.path within the context.

    Parameters
    ----------
    package : str
        package name
    version : Optional[str]
        package version, by default, latest version.
    validate_npe1_imports: bool
        Whether to try to import an npe1 plugin's entry points. by default True.
    install_napari_if_necessary: bool
        If `validate_npe1_imports` is True, whether to install napari if the import
        fails.  (It's not uncommon for plugins to fail to specify napari as a
        dependency.  Othertimes, they simply need a qt backend.).  by default True.

    Yields
    ------
    build.env.IsolatedEnv
        env object that has an `install` method.
    """
    with IsolatedEnvBuilder() as env:
        # install the package
        pkg = f"{package}=={version}" if version else package
        logger.debug(f"installing {pkg} into virtual env")
        env.install([pkg])

        # temporarily add env site packages to path
        prefixes = [getattr(env, "path")]  # noqa
        if not (site_pkgs := site.getsitepackages(prefixes=prefixes)):
            raise ValueError("No site-packages found")  # pragma: no cover
        sys.path.insert(0, site_pkgs[0])
        try:
            if validate_npe1_imports:
                # try to import the plugin's entry points
                dist = metadata.distribution(package)
                ep_groups = {ep.group for ep in dist.entry_points}
                if NPE1_ENTRY_POINT in ep_groups and NPE2_ENTRY_POINT not in ep_groups:
                    try:
                        _get_loaded_mf_or_die(package)
                    except Exception:  # pragma: no cover
                        # if loading contributions fails, it can very often be fixed
                        # by installing `napari[all]` into the environment
                        if install_napari_if_necessary:
                            env.install(["napari[all]"])
                            # force reloading of qtpy
                            sys.modules.pop("qtpy", None)
                            _get_loaded_mf_or_die(package)
                        else:
                            raise
            yield env
        finally:
            # cleanup sys.path
            sys.path.pop(0)


def _get_loaded_mf_or_die(package: str) -> PluginManifest:
    """Return a fully loaded (if npe1) manifest, or raise an exception."""
    from npe2 import PluginManifest
    from npe2.manifest._npe1_adapter import NPE1Adapter

    mf = PluginManifest.from_distribution(package)
    if isinstance(mf, NPE1Adapter):
        with warnings.catch_warnings():
            warnings.filterwarnings("error", message="Error importing contributions")
            warnings.filterwarnings("error", message="Failed to convert")
            warnings.filterwarnings("ignore", message="Found a multi-layer writer")
            mf._load_contributions(save=False)
    return mf


def _fetch_manifest_with_full_install(
    package: str, version: Optional[str] = None
) -> PluginManifest:
    """Fetch manifest for plugin by installing into an isolated environment."""
    # create an isolated env in which to install npe1 plugin
    with isolated_plugin_env(
        package, version, validate_npe1_imports=True, install_napari_if_necessary=True
    ):
        return _get_loaded_mf_or_die(package)


@lru_cache
def get_hub_plugins() -> Dict[str, str]:
    """Return {name: latest_version} for all plugins on the hub."""
    with urlopen("https://api.napari-hub.org/plugins") as r:
        return json.load(r)


@lru_cache
def get_hub_plugin(plugin_name: str) -> Dict[str, Any]:
    """Return hub information for a specific plugin."""
    with urlopen(f"https://api.napari-hub.org/plugins/{plugin_name}") as r:
        return json.load(r)
